/*
 * RLSInvert.c
 *
 *  Created on: Apr 16, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) cora.nwra.com)
 */
#include <stdlib.h>
#include <string.h>
#include "matrix.h"
#include "averagingKernel.h"
#include "lambda.h"
#include "errorHandler.h"
#include "RLSErrorMessages.h"
#include "ProblemDimension.h"
#include "utilities.h"
#include "BaseDirectory.h"
#include "Model.h"
#include "TempData.h"
#include "linearEqSolve.h"

static MATRIX M;
static MATRIX dasMatrix;
static MATRIX R;
static MATRIX CI;
static MATRIX C;  // inverse of CI
static MATRIX V;
static MATRIX d3asMatrix;
static MATRIX MTCIM;
static MATRIX B;
static MATRIX MTCI;

/**
 * tdSetup calculates all the non-changing, static matrices
 * @param[in] aM Kernel matrix
 * @param[in] ad map data
 * @param[in] aR	Regularization matrix; symmetric positive-definite matrix
 * @param[in] aCI Inverse of matrix C, a noise covariance matrix
 * @param[in] aC matrix C, a noise covariance matrix
 * @ingroup Matrix
 */
int tdSetup1d(MATRIX aM, MATRIX ad, MATRIX aR, MATRIX aCI, MATRIX aC) {

	M = aM;
	dasMatrix = ad;
	R = aR;
	CI = aCI;
	C = aC;

	// calculate intermediate values that are not dependent on lambda
	MTCI = multiplyMM2('T', 'N', M, CI);
	MTCIM = multiplyMM2('N', 'N', MTCI, M);

	V = MTCI;
	d3asMatrix = multiplyMM2('N', 'N', MTCI, dasMatrix);

	B = newMatrix(R.nRow, R.nCol);

	return TRUE;
}


int tdSetup3d(MATRIX aM, MATRIX ad, MATRIX aR, MATRIX aCI, MATRIX aC) {

	M = aM;
	dasMatrix = ad;
	R = aR;
	CI = aCI;
	C = aC;

	// calculate intermediate values that are not dependent on lambda
	MTCI = multiplyMM2('T', 'N', M, CI);

	V = MTCI;
	d3asMatrix = multiplyMM2('N', 'N', MTCI, dasMatrix);

	return TRUE;
}

/** Completes the inversion calculation.
 *  By sending the individual lambda values the inversion calculation
 *  is finished.
 * @param[in]  lambda regularization parameter
 * @param[in] debug Flag to print intermediate matrices during inversion
 * @param[in] debugFilename Base directory and filename for intermediate, debug values.
 * @param[in,out] tempData TempData structure contains all calculated solutions including
 * averaging kernel and error covariance for one wave number.
 * @ingroup Matrix
 */
void tdinvert3dFinish(double lambda, TempData *tempData, BOOLEAN debug, BaseDirectory *debugFilename) {

	MATRIX A; // Averaging Kernel
	MATRIX Cov; // full error covariance of the output
	MATRIX basMatrix; // parameters to model. We are solving for Matrix, b;
	MATRIX F;

	//                             Bb = d
	// [MT * CI * M + lambda * R] * b = MT * CI * d

	if (lambda == 0.0)
		reportError(ZeroLambda, "", WARNING);

	B = multiplyABplusC('N', 'N', MTCI, M, lambda, R);

	if (debug) {
		BOOLEAN compress = FALSE;
		writefits2d(finishFilename(debugFilename, "B", compress, debug), MatrixToFitsObject(B));
		writefits2d(finishFilename(debugFilename, "d3asMatrix", compress, debug), MatrixToFitsObject(d3asMatrix));
		writefits2d(finishFilename(debugFilename, "MTCIM", compress, debug), MatrixToFitsObject(MTCIM));
		writefits2d(finishFilename(debugFilename, "MTCI", compress, debug), MatrixToFitsObject(MTCI));
		writefits2d(finishFilename(debugFilename, "CI", compress, debug), MatrixToFitsObject(CI));
		writefits2d(finishFilename(debugFilename, "d", compress, debug), MatrixToFitsObject(dasMatrix));
		writefits2d(finishFilename(debugFilename, "R", compress, debug), MatrixToFitsObject(R));
	}
	basMatrix = linearEqSolve(B, d3asMatrix);
	if (debug) {
		BOOLEAN compress = FALSE;
		writefits2d(finishFilename(debugFilename, "m", compress, debug), MatrixToFitsObject(basMatrix));
	}

	// Now, find the averaging kernel
	/*       T  -1              -1   T  -1
	 * F = [M  C   M + lambda R]    M  C
	 *
	 *
	 *           T  -1
	 * Let B = [M  C   M + lambda R]
	 *
	 *           T  -1
	 *  and V = M  C
	 *
	 *             -1
	 *  Then, F = B   V
	 *
	 *   or BF = V, then find F, by solving system of linear equations one k at a time
	 *   where k indexes through each column of F and V.
	 */


	if (calcAvgKernelCov()) {
		// solves A*X = B for X.  We are solving B*F = V for F
		// solve using the LUDecomp calculated from  linearEqSolve
		F = LinearEqSolveNRHS(B, V);

		// Finally, calculate the Avg Kernel & Error Covariance; subset as allowed
		//Cov = F * C * FT;
		MATRIX Temp;

		MATRIX FR;
		if (subsetKernels()) { // then only calculate needed values for the averaging kernels
			// pull rows of F!
			FR = selectKernels(F);
		} else { // calculate full Avg Kern & Cov
			FR = F;
		}
		// calculate averaging kernel
		A = multiplyMM2('N', 'N', FR, M);
		// calculate error covariance
		Temp = multiplyMM2('N', 'N', FR, C);
		Cov = multiplyMM2('N', 'T', Temp, F);
		if (subsetKernels()) freeMatrix(FR);
		freeMatrix(Temp);
	}
	MATRIX dtilda;
	// calculate d~
	dtilda = multiplyMM2('N', 'N', M, basMatrix);

	if (calcAvgKernelCov()) {
		freeMatrix(F);
	}

	tempData->A = A;
	tempData->Cov = Cov;
	tempData->dTilda = dtilda;
	tempData->m = basMatrix;
}

/** Completes the inversion calculation.
 *  By sending the individual lambda values the inversion calculation
 *  is finished.
 * @param[in]  lambda regularization parameter
 * @param[in] nStep Nth step in lambda-parameter range
 * @param[in] debug Flag to print intermediate matrices during inversion
 * @param[in] debugFilename Base directory and filename for intermediate, debug values.
 * @param[in,out] model Model structure contains all calculated solutions including
 * averaging kernel and error covariance.
 * @ingroup Matrix
 */
void tdinvert1dFinish(double lambda, int nStep, Model *model, BOOLEAN debug, BaseDirectory *debugFilename) {

	MATRIX A; // Averaging Kernel
	MATRIX Cov; // full error covariance of the output
	MATRIX basMatrix; // parameters to model. We are solving for Matrix, b;
	MATRIX F;

	//                             Bb = d
	// [MT * CI * M + lambda * R] * b = MT * CI * d

	if (lambda == 0.0)
		reportError(ZeroLambda, "", WARNING);

	APlusLambdaB(MTCIM, lambda, R, B);

	if (debug) {
		BOOLEAN compress = FALSE;
		writefits2d(finishFilename(debugFilename, "B", compress, debug), MatrixToFitsObject(B));
		writefits2d(finishFilename(debugFilename, "d3asMatrix", compress, debug), MatrixToFitsObject(d3asMatrix));
		writefits2d(finishFilename(debugFilename, "MTCIM", compress, debug), MatrixToFitsObject(MTCIM));
		writefits2d(finishFilename(debugFilename, "MTCI", compress, debug), MatrixToFitsObject(MTCI));
		writefits2d(finishFilename(debugFilename, "CI", compress, debug), MatrixToFitsObject(CI));
		writefits2d(finishFilename(debugFilename, "d", compress, debug), MatrixToFitsObject(dasMatrix));
		writefits2d(finishFilename(debugFilename, "R", compress, debug), MatrixToFitsObject(R));
	}
	basMatrix = linearEqSolve(B, d3asMatrix);

	// Now, find the averaging kernel
	/*       T  -1              -1   T  -1
	 * F = [M  C   M + lambda R]    M  C
	 *
	 *
	 *           T  -1
	 * Let B = [M  C   M + lambda R]
	 *
	 *           T  -1
	 *  and V = M  C
	 *
	 *             -1
	 *  Then, F = B   V
	 *
	 *   or BF = V, then find F, by solving system of linear equations one k at a time
	 *   where k indexes through each column of F and V.
	 */

	if (debug) {
		BOOLEAN compress = FALSE;
		writefits2d(finishFilename(debugFilename, "m", compress, debug), MatrixToFitsObject(basMatrix));
	}
	// solves A*X = B for X.  We are solving B*F = V for F
	// solve using the LUDecomp calculated from  linearEqSolve
	F = LinearEqSolveNRHS(B, V);

	// Finally, calculate the Avg Kernel & Error Covariance; subset as allowed
	//Cov = F * C * FT;
	MATRIX Temp;

	// calculate averaging kernel
	A = multiplyMM2('N', 'N', F, M);
	// calculate error covariance
	Temp = multiplyMM2('N', 'N', F, C);
	Cov = multiplyMM2('N', 'T', Temp, F);
	freeMatrix(Temp);

	freeMatrix(F);

	model->A1d = A;
	model->Cov1d = Cov;
	saveB(model, basMatrix, nStep);
	freeMatrix(basMatrix);
}


/**
 * free memory used by static variables internal to inversion
 * @ingroup Matrix
 */
void tdCleanup() {

	// free intermediate values that are not dependent on lambda
	freeMatrix(MTCI);
	if (getProblemDimension() == 1) {
		freeMatrix(MTCIM);
		freeMatrix(CI);
	}
	freeMatrix(d3asMatrix);
	freeMatrix(B);

}
