/*
 * main.c
 *
 *  Created on: Aug 16, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) cora.nwra.com)
 */

#include <string.h>
#include <math.h>
#include "utilities.h"
#include "Constants.h"
#include "errorHandler.h"
#include "coramfVersion.h"
#include "parsVersion.h"
#include "invrsVersion.h"
#include "Inversion.h"
#include "RLSInit.h"
#include "RLSInvert.h"

#include "ProblemDimension.h"

/**
 * \page InvTypesPage
 * <B>Regularized least squares inversion (RLS) </B>
 *
 * @see <a href="../RLSnotes.pdf">Equations and specifics</a>
 * @see RLSInversion
 * @see RLSMain.c
 */


/**
 * Performs RLS 1D, 3D scalar and 3D vector inversions.
 * There is
 * a distinction  between the 1D and 3D problems in terms of the input
 * data.
 * The 1D problem uses 2D matrices and the solution is complete
 * after each inversion.  The 3D problem uses 4D matrices and
 * partial solutions are produced by each inversion.  The partial
 * solutions are aggregrated into full 4D solutions and written
 * when they are complete.
 * @param[in] inputParms InputParms structure which contains input file names, directories, and
 * parameters for the inversion.
 * @param[in] commandLineOutputDir Character string to override the output base
 * directory specified in the input file.  This parameter is used to create successive
 * versions of the output base directory.
 * @param[in] debug Boolean flag signaling the output of intermediate matrix values.
 * @returns 0 for success and 1 for error.
 */
int run(InputParms inputParms, char *commandLineOutputDir, BOOLEAN debug) {

	static struct Inversion Rls3d = {
			NULL, // InputMatrices
			// define function pointers and initialize to NULL
			&RLSInit3d,
			&RLSInvert3d,
			&RLSComplete3d,
			&RLSDestroy3d,
			&RLSMoreParameterSteps,
			RLSNextParameter3d
	};

	static struct Inversion Rls1d = {
			NULL, // InputMatrices
			// define function pointers and initialize to NULL
			&RLSInit1d,
			&RLSInvert1d,
			&RLSComplete1d,
			&RLSDestroy1d,
			&RLSMoreParameterSteps,
			&RLSNextParameter1d
	};

	printf("Inversion %s\n", CODE_VERSION);
	printf("Libraries %s\n", getLibraryVersion());
	printf("coramf Version %s\n", coramfVersion);
	printf("pars Version %s\n", parsVersion);
	printf("invrs Version %s\n", invrsVersion);
	// check library and inversion for compatible versions
	if ((strcmp(CODE_VERSION, getLibraryVersion()) != 0) ||
			(strcmp(CODE_VERSION, coramfVersion) != 0) ||
			(strcmp(CODE_VERSION, parsVersion) != 0) ||
			(strcmp(CODE_VERSION, invrsVersion) != 0))
		reportError(BadCodeVersions, "main", FATAL);

	struct Inversion op;

	if (inputParms.rls) {
		if (getProblemDimension() == PDC.D1) {
			op = Rls1d;
		} else {
			op = Rls3d;
		}
		printf("using rls\n");
	} else
		reportError("Unrecognized type of inversion", "main", FATAL);

	op.data = op.init(inputParms, commandLineOutputDir, debug);

	// loop over parameters
	while (op.moreParameters(op.data)) {
		op.solve(op.data);
		op.complete(op.data);
		op.nextParameter(op.data);
	}
	// end loop
	op.destroy(op.data);

	return(EXIT_SUCCESS);

}

char *getRLSVersion() {
	return CODE_VERSION;
}

